#!/usr/bin/perl -w

use strict;
use Test::More tests => 39;
use Data::Dumper;

# 1
BEGIN { use_ok('FDN::Common::Config'); }

$FDN::Common::Config::CONFIG_ROOT = './t/';
$FDN::Common::Config::debug=1;

sub clean_tmp_files() {
    # n
    if ( -f 't/tmp.cfg' ) {
        system("rm -f t/tmp.cfg");
    }
    ok(! -f 't/tmp.cfg',	'Temporary config file correctly deleted');

    # n+1
    if ( -f 't/tmp.cfg.bin' ) {
        system("rm -f t/tmp.cfg.bin");
    }
    ok(! -f 't/tmp.cfg.bin',	'Temporary config (bin) file correctly deleted');
    
}

sub set_tmp_file($) {
    my ($text) = @_;
    open(F, ">t/tmp.cfg");
    print F $text;
    close F;
    # n
    ok(-f 't/tmp.cfg',		'Temporary config file correctly created');
}


my $file1 = 
	";\n".
	"[test]\n".
	"key = val\n".
	"\n".
	"[retest]\n".
	"rekey = reval\n";

my $file2 = 
	";\n".
	"[test]\n".
	"key = newval\n".
	"\n".
	"[retestnew]\n".
	"rekey = renewval\n";

my $cfg;
my $tmp;
my $tmp2;

# 2
$cfg = Load_Config('dummy.cfg');
is($cfg, undef, 	 	'Non-existent config file');

if ( -f './t/test.cfg.bin' ) {
    system("rm ./t/test.cfg.bin");
}

# 3, 4
clean_tmp_files();

# 5
set_tmp_file($file1);

$FDN::Common::Config::CONFIG_ROOT = './t/';

# 6
$cfg = Load_Config('test.cfg');
ok(defined $cfg, 		'Loaded test.cfg config file');

# 7
is(ref $cfg, 'HASH', 		'$cfg is a HASH reference');

# 8
is($cfg->{'Section1'}->{'clef28'}, 'Valeur 28',
				q($cfg->{'Section1'}->{'clef28'} correctly set));
# 9
$tmp = Config_Sections_List('test.cfg');
is(ref $tmp, 'ARRAY',		'$tmp (section list) is a ARRAY reference');

# 10
ok(eq_set($tmp,['Section1','Section3','SectionI']),
				'Section list is good');

# 11
$tmp = Config_Section('test.cfg','Section1');
is(ref $tmp, 'HASH',		'$tmp (Section1) is a HASH reference');

# 12
is($tmp->{'clef21'}, 'Valeur 21',
				'clef 21 in $tmp is good');

# 13
$tmp = Config_Section_Keys('test.cfg','Section2');
is($tmp, undef,			'There is no section Section2');

# 14
$tmp = Config_Section_Keys('test.cfg','Section3');
is(ref $tmp, 'ARRAY',		'$tmp (Section3 keys) is a ARRAY reference');

# 15
ok(eq_set($tmp,['clef1','clef2','clef3','@clef4','@clef5']),
				'$tmp (Section3 keys) is good');

# 16
$tmp = Config_Key('test.cfg','Section3','clef1');
is($tmp, 'Valeur 3.1-2',	'key override cross-boundaries is ok');

# 17
$tmp = Config_Key('test.cfg','Section3','clef2');
is($tmp, 'Valeur 3.2-1',	'key cross-boundaries is ok');

# 18
$cfg = Load_Config('tmp.cfg');
ok(defined $cfg,		'Temporary config file loaded');

# 19
ok(-f 't/tmp.cfg.bin',		'Temporary config (bin) file created');

# 20
is(ref $cfg, 'HASH',		'$cfg is a HASH reference');

# 21
is($cfg->{'test'}->{'key'}, 'val',
				'$cfg->{\'test\'}->{\'key\'} correctly set');

# 22
is($cfg->{'retest'}->{'key'}, undef,
				'$cfg->{\'retest\'}->{\'key\'} correctly not set');

# 23
$tmp = Config_Key('test.cfg','Section3','@clef4');
ok(ref($tmp) eq 'ARRAY',	'Multiple key is an array');

# 24
ok(eq_set($tmp,['Valeur 3.4.1','Valeur 3.4.2']),
				'Multiple key values are good');

# 25
$tmp = Config_Key('test.cfg','Section3','@clef5');
ok(ref($tmp) eq 'ARRAY',	'Multiple key (with single value) is an array');

# 26
ok(eq_set($tmp,['Valeur 3.5.1']),
				'Multiple key values are good');


# 27
$tmp = Config_Key('test.cfg','Section1','clef20');
is($tmp, 'Valeur 1, blah',      'Reference interne simple directe');

# 28
$tmp = Config_Key('test.cfg','SectionI','clef1');
is($tmp, 'Valeur 1, blah',	'Reference interne simple indirecte');

# 29
$tmp = Config_Key('test.cfg','SectionI','clef2');
is($tmp, 'Valeur 1, blah',	'Reference interne simple indirecte');

# 30
$tmp = Config_Key('test.cfg','SectionI','clef6');
is($tmp, 'debut, fin',		'Protection virgule et cat');

# 31
$tmp = Config_Key('test.cfg','SectionI','clef7');
is($tmp, 'La valeur, finalement, est Valeur 1, blah', 'Protection virgule complexe');

# 32
$tmp = Config_Key('test.cfg','SectionI','clef8');
is($tmp, '(un essai, a la con)',	'Protection parenthese et virgule');

# 33
$tmp = Config_Key('test.cfg','SectionI','clef10');
is($tmp, 'val',			'External reference');

# 34
$tmp = Config_Key('test.cfg','SectionI','@clef11');
ok(eq_set($tmp,['Valeur 3.4.1','Valeur 3.4.2','Valeur 3.4.1','Valeur 3.4.2']),
				'Multiple key values in references');

[ -f './t/tmp2.cfg' ] && system('rm -f ./t/tmp2.cfg');
[ -f './t/tmp2.cfg.bin' ] && system('rm -f ./t/tmp2.cfg.bin');

# 35
$tmp = Write_Config('./t/tmp2.cfg',{'sect'=>{'clef'=>'val'}});
ok ( !$tmp, 'Write_Config');

# 36
$tmp = Load_Config('tmp2.cfg');
ok(ref($tmp) eq 'HASH', 'Lecture nouveau fichier');

# 37
ok(defined $tmp->{'sect'}, 'Lecture nouveau fichier');

# 38
ok(defined $tmp->{'sect'}->{'clef'}, 'Lecture nouveau fichier');

# 39
is($tmp->{'sect'}->{'clef'}, 'val', 'Lecture nouveau fichier');



#
# Just to make vim's color syntax engine know that this '.t' file is a Perl
# script :
#
# vim:se filetype=perl:
#
