#!/usr/bin/perl

use strict;
use Test::More tests => 476;
use FDN::Common::Logger;
use FDN::Common::Constantes;
#use FDN::Test::Workflow;
use English;

# 1
BEGIN { use_ok('FDN::Sql::DB_Access') }

# 2
BEGIN { use_ok('FDN::Sql::Basic') }

# 3
BEGIN { use_ok('FDN::Sql::DBC') }

# 4
BEGIN { use_ok('FDN::Sql::Workflow::Basic') }

# 5
BEGIN { use_ok('FDN::Common::Config') }
BEGIN { $FDN::Common::Config::CONFIG_ROOT = './t/' }

# 6
BEGIN { use_ok('FDN::Sql::Workflow::Run') }

# 7
my $dbh = do_connect('test');
ok ( defined $dbh, 'do_connect' );

# 8
is ( ref($dbh), 'HASH', 'do_connect' );

# 9
ok ( -f './sql/workflow.sql', 'workflow.sql' );
open FILE, '<./sql/workflow.sql';
my $workflow;
{ local $/;
  $workflow = <FILE>; }
close FILE;
my @queries = split /;\s*$/m, $workflow;
my $tmp;
my $res = TRUE;
for my $q ( @queries ) {
	next if ( $q =~ /drop table if exists SEQ/ );
	next if ( $q =~ /create table SEQ/ );
	$tmp = sql_transaction($dbh,'create',$q);
	if ( ! $tmp->{'ok'} ) {
		$res = FALSE;
		print "# Failed query: $q\n";
	}
}
# 10
ok ( $res, 'workflow.sql' );

# 11
$res = workflow_select_db('test');
is ( $res, $dbh, 'workflow_select_db' );

####
#### Basic testing ->
####

# /1,3
$res = get_task_list();
is ( ref($res), 'ARRAY', 'get_task_list' );
is ( $#$res, 3, 'get_task_list' );
ok ( eq_set( $res, ['essai', 'complexe', 'serie', 'para' ] ), 'get_task_list' );

# /4
$res = get_task_infos('autre');
is ( $res, undef, 'get_task_infos' );

# /5
my $in1 = {
	opt => 'mandatory',
	syn => '.*',
	name => 'nom'
};
my $in2 = {
	opt => 'optional',
	syn => '\d+',
	name => 'telephone'
};
my $out1 = {
	opt => 'mandatory',
	syn => '(oui|non)',
	name => 'etat'
};
my $infos = {
	'type' => 'atomic',
	'keycalc' => 'CAT',
	'action' => \&FDN::Test::Workflow::faire_un_essai,
	'name' => 'essai',
	'in' => [ $in1, $in2 ],
	'out' => [ $out1 ],
	'byname' => {
		'nom' => $in1,
		'telephone' => $in2,
		'etat' => $out1
	},
	'launch' => []
};
$res = get_task_infos('essai');
is_deeply( $res, $infos, 'get_task_infos' );

# /6
$in2->{'syn'} = '^0[1-6]\d+$';
$infos = {
	'type' => 'atomic',
	'name' => 'complexe',
	'action' => \&FDN::Test::Workflow::faire_un_essai,
	'keycalc' => \&FDN::Test::Workflow::calculer_la_clef,
	'in' => [ $in1, $in2 ],
	'out' => [ $out1 ],
	'byname' => {
		'nom' => $in1,
		'telephone' => $in2,
		'etat' => $out1
	},
	'launch' => []
};
$res = get_task_infos('complexe');
is_deeply( $res, $infos, 'get_task_infos' );

# /7
$in2->{'syn'} = '\d+';
$out1->{'syn'} = '(oui|non)(oui|non)';
$infos = {
	'type' => 'serial',
	'keycalc' => 'CAT',
	'name' => 'serie',
	'action' => \&FDN::Test::Workflow::faire_un_essai,
	'in' => [ $in1, $in2 ],
	'out' => [ $out1 ],
	'byname' => {
		'nom' => $in1,
		'telephone' => $in2,
		'etat' => $out1
	},
	'launch' => [ {
		'name' => 'essai',
		'in' => {
			'nom' => '$nom',
			'telephone' => \&FDN::Test::Workflow::faire_telephone
		},
		'out' => {
			'etat' => 'essai-etat'
		}
	}, {	
		'name' => 'complexe',
		'in' => {
			'nom' => 'nom',
			'telephone' => '$telephone'
		},
		'out' => {
			'etat' => 'complexe-etat'
		}
	} ]
};
$res = get_task_infos('serie');
is_deeply( $res, $infos, 'get_task_infos' );

# /8
$in1 = {
	name => 'serial',
	opt => 'mandatory',
	syn => '.*'
};
$infos = {
	'type' => 'parallel',
	'name' => 'para',
	'keycalc' => 'CAT',
	'action' => \&FDN::Test::Workflow::faire_un_essai,
	'in' => [ $in1 ],
	'out' => [],
	'byname' => { 'serial' => $in1 },
	'launch' => [ {
		name => 'serie',
		in => {
			'nom' => '$serial',
			'telephone' => '0100112233'
		},
		out => {}
	}, {
		name => 'complexe',
		in => {
			'nom' => '$serial',
			'telephone' => '0122'
		},
		out => {}
	} ]
};
$res = get_task_infos('para');
is_deeply( $res, $infos, 'get_task_infos' );

# /9
my $test;
my $vars;
my $props;
my $props_pere;
my $props_fils;
my $exp;
$test = \&FDN::Sql::Workflow::Run::check_start_vars;
$infos = get_task_infos('essai');
$vars = {
	'nom' => 'Pas de contrainte',
	'telephone' => 'Digits only'
};
$res = &$test('essai',$infos,$vars);
is ( $res, FALSE, 'check_start_vars' );

# /10
$vars = {
	'nom' => 'Pas de contrainte',
	'telephone' => '123'
};
$res = &$test('essai',$infos,$vars);
is ( $res, 'essai/Pas de contrainte', 'check_start_vars' );

# /11
$infos = get_task_infos('complexe');
$vars = {
	'nom' => 'Pas de contrainte',
	'telephone' => '01000'
};
$res = &$test('complexe',$infos,$vars);
is ( $res, 'Pas de contrainte-01000!complexe', 'check_start_vars' );

# /12
# 'nom' is taken from daddy's props and
# 'telephone' is taken from the function 'faire_telephone'
$test = \&FDN::Sql::Workflow::Run::prepare_entree_fils;
$props = {
	'nom' => 'Nom',
	'telephone' => '0102'
};
$infos = get_task_infos('serie');
$exp = {
	'nom' => 'Nom',
	'telephone' => FDN::Test::Workflow::faire_telephone(undef)
};
$res = &$test($infos,$props,0);
is_deeply( $res, $exp, 'prepare_entree_fils' );

# /13
# 'nom' is a constant given in the config file and
# 'telephone' is taken from daddy's props.
$exp = {
	'nom' => 'nom',
	'telephone' => '0102'
};
$res = &$test($infos,$props,1);
is_deeply( $res, $exp, 'prepare_entree_fils' );

# /14,15
$test = \&FDN::Sql::Workflow::Run::traite_sortie_fils;
$exp = {
	'inutile' => 'useless',
	'nom' => 'Un nom',
	'telephone' => '0000'
};
$props_fils = {
	'nom' => 'Other name useless',
	'telephone' => 'Anything not used',
	'etat' => 'oui'
};
$props_pere = $exp;
# The propagation is supposed to add that property:
$exp->{'essai-etat'} = 'oui';
# As it adds a property, the result is not clean (needs an update in DB).
$res = &$test($infos,$props_pere,$props_fils,0);
is ( $res, FALSE, 'traite_sortie_fils' );
is_deeply( $props_pere, $exp, 'traite_sortie_fils' );

# /16,17
$props_fils = {
	'nom' => 'rien',
	'etat' => 'non'
};
$props_pere = $exp;
$exp->{'complexe-etat'} = 'non';
$res = &$test($infos,$props_pere,$props_fils,1);
is ( $res, FALSE, 'traite_sortie_fils' );
is_deeply( $props_pere, $exp, 'traite_sortie_fils' );

# Before 11 + (/17) -> 28

####
#### Higher level functions -> 168
####
error "NO MORE ERRORS.";
my ( $tid, @tid );
my ( $key, @key );

# launch_task est assez simple: elle cree la tache, lui attache les
# proprietes qui vont bien, et cree le premier RUN sur l'evennement
# 'init'.

# /1,6
( $tid[0], $key ) = launch_task('essai', { nom => "Un nom", telephone => '010203' } );
ok ( $tid[0] > 0, 'launch_task' );
is ( $key, 'essai/Un nom', 'launch_task' );
$res = task_par_id($tid[0]);
is ( $res->{'TASK_KEY'}, $key, 'launch_task' );
is ( $res->{'TASK_STATE'}, 'init', 'launch_task' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 0, 'launch_task' );
is ( $res->[0]->{'RUN_EVENT'}, 'init', 'launch_task' );

# __fail_task est aussi facile a tester: pose la tache a l'etat 'failed'
# et annule toutes les executions a venir.
my $fail_task = \&FDN::Sql::Workflow::Run::__fail_task;

# /7,8
&$fail_task($tid[0]);
$res = task_par_id($tid[0]);
is ( $res->{'TASK_STATE'}, 'failed', '__fail_task' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__fail_task' );
# FIXME S'il y avait un waiting, et si l'evennement survient de maniere suffisament
# proche, il y a un *vrai* risque que la tache soit encore jouee une fois.
# Pour empecher ca, il faudrait locker les evennement pendant qu'une tache se joue,
# de maniere a ce qu'un evennement ne puisse pas survenir entre sa demande d'echec
# et son echec effectif.

# __identification_tache, facile a tester: ajoute les clefs 'tid' et 'ptid'
# aux proprietes de la tache si elles n'y sont pas encore.
my $identification_tache = \&FDN::Sql::Workflow::Run::__identification_tache;

# /9,12
# Pas de pere connu.
$props = get_task_properties($tid[0]);
is ( $props->{'tid'},  undef, '__identification_tache' );
is ( $props->{'ptid'}, undef, '__identification_tache' );
&$identification_tache($tid[0],$props);
is ( $props->{'tid'}, $tid[0], '__identification_tache' );
is ( $props->{'ptid'}, -1, '__identification_tache' );

# /13,16
# On lui cree un pere, et on test avec un pere.
$tid[1] = launch_task('serie', { nom => "Un nom", telephone => '010203' } );
&$fail_task($tid[1]);
new_taskself($tid[1],$tid[0],1);
$props = get_task_properties($tid[0]);
is ( $props->{'tid'},  undef, '__identification_tache' );
is ( $props->{'ptid'}, undef, '__identification_tache' );
&$identification_tache($tid[0],$props);
is ( $props->{'tid'},  $tid[0], '__identification_tache');
is ( $props->{'ptid'}, $tid[1], '__identification_tache');

# __traite_sortie_tache, facile a tester: si la tache retourne undef, elle
# echoue, si elle retourne vrai c'est un evennement attendu, si elle retourne
# faux, rien de special a faire.
my $traite_sortie_tache = \&FDN::Sql::Workflow::Run::__traite_sortie_tache;

# /17,22
# Sortie en echec
set_task_state($tid[0],'running');
new_run($tid[0],'fake1');
new_run($tid[0],'fake2');
new_waiting($tid[0],'evout1');
new_waiting($tid[0],'evout2');
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, 1, '__traite_sortie_tache echec' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 1, '__traite_sortie_tache echec' );
$res = &$traite_sortie_tache($tid[0],undef);
ok ( ! $res, '__traite_sortie_tache echec' );
$res = task_par_id($tid[0]);
is ( $res->{'TASK_STATE'}, 'failed', '__traite_sortie_tache echec' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache echec' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache echec' );

# /23,28
# Sortie en succes
set_task_state($tid[0],'running');
new_run($tid[0],'fake1');
new_run($tid[0],'fake2');
new_waiting($tid[0],'evout1');
new_waiting($tid[0],'evout2');
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, 1, '__traite_sortie_tache succes' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 1, '__traite_sortie_tache succes' );
$res = &$traite_sortie_tache($tid[0],0);
ok ( $res, '__traite_sortie_tache succes' );
$res = task_par_id($tid[0]);
is ( $res->{'TASK_STATE'}, 'running', '__traite_sortie_tache succes' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, 1, '__traite_sortie_tache succes' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 1, '__traite_sortie_tache succes' );

# /29,36
# Sortie evennement en attente
&$fail_task($tid[0]);
set_task_state($tid[0],'running');
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache waiting' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache waiting' );
$res = &$traite_sortie_tache($tid[0],'evout1');
ok ( $res, '__traite_sortie_tache waiting' );
$res = task_par_id($tid[0]);
is ( $res->{'TASK_STATE'}, 'running', '__traite_sortie_tache waiting' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, 0, '__traite_sortie_tache waiting' );
is ( $res->[0]->{'WAITING_EVENT'}, 'evout1', '__traite_sortie_tache waiting' );
is ( $res->[0]->{'TASK_ID'}, $tid[0], '__traite_sortie_tache waiting' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache waiting' );

# /37,44
# Sortie evennement en survenu
&$fail_task($tid[0]);
send_event('evout1');
set_task_state($tid[0],'running');
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache running' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache running' );
$res = &$traite_sortie_tache($tid[0],'evout1');
ok ( $res, '__traite_sortie_tache running' );
$res = task_par_id($tid[0]);
is ( $res->{'TASK_STATE'}, 'running', '__traite_sortie_tache running' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__traite_sortie_tache running' );
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 0, '__traite_sortie_tache running' );
is ( $res->[0]->{'RUN_EVENT'}, 'evout1', '__traite_sortie_tache running' );
is ( $res->[0]->{'TASK_ID'}, $tid[0], '__traite_sortie_tache running' );

# __find_last_son: doit nous retourner le numero et l'id du dernier fils
# connu d'une tache donnee. L'indication qu'on donne est une hypothese
# heuristique: ca pourrait etre ca (cas normal), sauf si des fils ont
# disparu entre temps (il y a eu un echec, on a annule les derniers fils,
# et on a relance).
my $find_last_son = \&FDN::Sql::Workflow::Run::__find_last_son;

# /45,46
( $res, $tid ) = &$find_last_son($tid[1],3);
is ( $tid, $tid[0], '__find_last_son' );
is ( $res, 1, '__find_last_son' );

# /47,48
( $res, $tid ) = &$find_last_son($tid[1],1);
is ( $tid, $tid[0], '__find_last_son' );
is ( $res, 1, '__find_last_son' );

# /49,50
del_taskself($tid[1],$tid[0]);
( $res, $tid ) = &$find_last_son($tid[1],5);
is ( $res, 0, '__find_last_son' );
is ( $tid, -1, '__find_last_son' );
new_taskself($tid[1],$tid[0],1);

# __propage_fils: quand un fils se termine, les infos utiles sont remontees
# vers le pere comme l'indique la configuration.
#
# Pour le test, on a deja le pere $tid[1], il a deja un premier fils $tid[0].
# On va lui creer un deuxieme fils $tid[2], et on va leur ajouter les infos
# qu'ils sont censes remonter. On va bien verifier que tout ce monde-la soit
# echoue avant de travailler (ca evitera les parasites par la suite).
my $propage_fils = \&FDN::Sql::Workflow::Run::__propage_fils;
$tid[2] = launch_task('complexe', {nom => 'nom', telephone => '020000'} );
new_taskself($tid[1],$tid[2],2);
&$fail_task($tid[0]);
&$fail_task($tid[1]);
&$fail_task($tid[2]);
$res = get_task_properties($tid[0]);
$res->{'etat'} = 'oui';
update_task_properties($tid[0],$res);
$res = get_task_properties($tid[2]);
$res->{'etat'} = 'non';
update_task_properties($tid[2],$res);
$infos = get_task_infos('serie');

# /51,56
$res = get_task_properties($tid[1]);
is ( $res->{'essai-etat'}, undef, '__propage_fils' );
is ( $res->{'complexe-etat'}, undef, '__propage_fils' );
$res = &$propage_fils($tid[1],$infos,'EOT-essai/Un nom');
ok ( $res, '__propage_fils' );
$res = get_task_properties($tid[1]);
is ( $res->{'essai-etat'}, 'oui', '__propage_fils' );
$res = &$propage_fils($tid[1],$infos,'EOT-nom-020000!complexe');
ok ( $res, '__propage_fils' );
$res = get_task_properties($tid[1]);
is ( $res->{'complexe-etat'}, 'non', '__propage_fils' );

# __try_end_task: si la tache n'attend plus rien (pas de WAITING et pas de RUN)
# alors on lance un RUN sur l'evennement 'end'.
#
# On va tester ca sur $tid[0].
my $try_end_task = \&FDN::Sql::Workflow::Run::__try_end_task;

# /57,61
# Rien en attente
&$fail_task($tid[0]);
set_task_state($tid[0],'running');
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__try_end_task cas 1' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__try_end_task cas 1' );
&$try_end_task($tid[0]);
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 0, '__try_end_task cas 1' );
is ( $res->[0]->{'RUN_EVENT'}, 'end', '__try_end_task cas 1' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__try_end_task cas 1' );

# /62,66
# Un run en attente
&$fail_task($tid[0]);
set_task_state($tid[0],'running');
new_run($tid[0],'event');
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 0, '__try_end_task cas 2' );
$res = [waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__try_end_task cas 2' );
&$try_end_task($tid[0]);
$res = [ run_par_task($tid[0]) ];
is ( $#$res, 0, '__try_end_task cas 2' );
is ( $res->[0]->{'RUN_EVENT'}, 'event', '__try_end_task cas 2' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, -1, '__try_end_task cas 2' );

# /67,71
# Un waiting en attente
&$fail_task($tid[0]);
set_task_state($tid[0],'running');
new_waiting($tid[0],'pasencore');
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__try_end_task cas 3' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, 0, '__try_end_task cas 3' );
&$try_end_task($tid[0]);
$res = [ run_par_task($tid[0]) ];
is ( $#$res, -1, '__try_end_task cas 3' );
$res = [ waiting_par_task($tid[0]) ];
is ( $#$res, 0, '__try_end_task cas 3' );
is ( $res->[0]->{'WAITING_EVENT'}, 'pasencore', '__try_end_task cas 3' );


# __launch_subtask: lance un des fils d'une tache. Retourne l'id de la
# nouvelle tache.
my $launch_subtask = \&FDN::Sql::Workflow::Run::__launch_subtask;
# On prepare le pere:
$props = {
	nom => 'Autre nom',
	telephone => '030405'
};
$infos = get_task_infos('serie');
$tid[3] = launch_task('serie',$props);

# /72,79
# Lancement du premier fils
$tid[4] = &$launch_subtask($tid[3],$infos,$props,0);
$res = task_par_id($tid[4]);
is ( $res->{'TASK_TYPE'}, 'atomic', '__launch_subtask' );
is ( $res->{'TASK_NAME'}, 'essai', '__launch_subtask' );
is ( $res->{'TASK_KEY'}, 'essai/Autre nom', '__launch_subtask' );
$res = [ run_par_task($tid[4]) ];
is ( $#$res, 0, '__launch_subtask' );
is ( $res->[0]->{'RUN_EVENT'}, 'init', '__launch_subtask' );
$res = [ waiting_par_task($tid[4]) ];
is ( $#$res, -1, '__launch_subtask' );
$res = [ waiting_par_task($tid[3]) ];
is ( $#$res, 0, '__launch_subtask' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-essai/Autre nom', '__launch_subtask' );

# /80,86
# Lancement du deuxieme fils
$tid[5] = &$launch_subtask($tid[3],$infos,$props,1);
$res = task_par_id($tid[5]);
is ( $res->{'TASK_TYPE'}, 'atomic', '__launch_subtask' );
is ( $res->{'TASK_NAME'}, 'complexe', '__launch_subtask' );
is ( $res->{'TASK_KEY'}, 'nom-030405!complexe', '__launch_subtask' );
$res = [ run_par_task($tid[5]) ];
is ( $#$res, 0, '__launch_subtask' );
is ( $res->[0]->{'RUN_EVENT'}, 'init', '__launch_subtask' );
$res = [ waiting_par_task($tid[5]) ];
is ( $#$res, -1, '__launch_subtask' );
$res = [ waiting_par_task($tid[3]) ];
is ( $#$res, 1, '__launch_subtask' );


# __post_run_serial, complique a tester. Se charge de lancer le prochain fils d'une tache
# serie si c'est necessaire. Si on est arrive a bout de tous les fils, essaye de terminer
# la tache. Si on est sur un evennement qui n'est pas la fin d'un fils et qu'il reste un
# fils a attendre, ne fait rien.
my $post_run_serial = \&FDN::Sql::Workflow::Run::__post_run_serial;
my $task;

# On va preparer le papa de tout ce beau monde, une vraie tache de type 'serial'
# toute neuve. On supprime le RUN sur 'init' et on passe la tache a running.
$props = {
	nom => 'Encore',
	telephone => '040506'
};
$infos = get_task_infos('serie');
$tid[6] = launch_task('serie',$props);
$task = task_par_id($tid[6]);
$res = run_par_task($tid[6]);
del_run($res->{'RUN_ID'});
set_task_state($tid[6],'running');

# /87,91
# Premier cas: on recoit 'init'. Normalement le premier fils a du etre lance, et on
# doit attendre sa fin.
$task = task_par_id($tid[6]);
&$post_run_serial($task, $infos, $props, 'init' );
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 1' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-essai/Encore', '__post_run_serial cas 1' );
$res = task_par_key('essai/Encore');
is ( ref($res), 'HASH', '__post_run_serial cas 1' );
$tid[7] = $res->{'TASK_ID'};
( $res, $tid ) = &$find_last_son($tid[6],5);
is ( $res, 1, '__post_run_serial cas 1' );
is ( $tid, $tid[7], '__post_run_serial cas 1' );

# /92,95
# Deuxieme cas:
# On recoit un evennement quelconque, alors qu'on attend encore au moins un fils. Il
# ne doit rien se passer.
$task = task_par_id($tid[6]);
&$post_run_serial($task, $infos, $props, 'quelconque' );
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 2' );
$res = [ run_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 2' );
( $res, $tid ) = &$find_last_son($tid[6],5);
is ( $res, 1, '__post_run_serial cas 2' );
is ( $tid, $tid[7], '__post_run_serial cas 2' );

# /96,106
# Troisieme cas: le premier fils se termine. Pour ca on envoie l'evennement, puis
# on annule le RUN correspondant, on passe egalement de fils dans l'etat d'une fin
# heureuse, puis on fait le post-traitement pour verifier.
# Normalement, le deuxieme fils doit se lancer.
send_event('EOT-essai/Encore');
$res = [ run_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 3' );
is ( $res->[0]->{'RUN_EVENT'}, 'EOT-essai/Encore', '__post_run_serial cas 3' );
del_run($res->[0]->{'RUN_ID'});
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 3' );
set_task_state($tid[7],'done');
$res = [ run_par_task($tid[7]) ];
is ( $#$res, 0, '__post_run_serial cas 3' );
is ( $res->[0]->{'RUN_EVENT'}, 'init', '__post_run_serial cas 3' );
del_run($res->[0]->{'RUN_ID'});
$task = task_par_id($tid[6]);
&$post_run_serial($task, $infos, $props, 'EOT-essai/Encore' );
$res = [ run_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 3' );
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 3' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-nom-040506!complexe', '__post_run_serial cas 3' );
$res = task_par_key('nom-040506!complexe');
is ( ref($res), 'HASH', '__post_run_serial cas 3' );
$tid[8] = $res->{'TASK_ID'};
( $res, $tid ) = &$find_last_son($tid[6],5);
is ( $res, 2, '__post_run_serial cas 3' );
is ( $tid, $tid[8], '__post_run_serial cas 3' );

# /107,109
# Quatrieme cas: je recois la fin d'une tache que je ne connais pas. Rien de particulier
# a faire. La tache en question n'est pas declaree, ca log une erreur.
$task = task_par_id($tid[6]);
&$post_run_serial($task, $infos, $props, 'EOT-inconnu' );
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 4' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-nom-040506!complexe', '__post_run_serial cas 4' );
$res = [ run_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 4' );

# /110,113
# Cinquieme cas: mon deuxieme fils part en couille. Il est purement et simplement annule,
# comme suite a un echec avec reprise manuelle. Je recupere un RUN sur FEOT du fils precedent.
# Je recois un evennement sans rapport. Il ne doit rien se passer.
&$fail_task($tid[8]);
set_task_key($tid[8],'failure-'.$tid[8]);
del_taskself($tid[6],$tid[8]);
new_taskself($tid[6],$tid[8],-1);
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 5' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-nom-040506!complexe', '__post_run_serial cas 5' );
del_waiting($res->[0]->{'WAITING_ID'});
new_run($tid[6],'FEOT-essai/Encore');
$task = task_par_id($tid[6]);
&$post_run_serial($task,$infos,$props,'sans-rapport');
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 5' );
$res = [ run_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 5' );

# /114,116
# Sixieme cas: on recoit le FEOT-* attendu. On doit de nouveau lancer le deuxieme fils.
$res = [ run_par_task($tid[6]) ];
del_run($res->[0]->{'RUN_ID'});
$task = task_par_id($tid[6]);
&$post_run_serial($task,$infos,$props,'FEOT-essai/Encore');
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 6' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-nom-040506!complexe', '__post_run_serial cas 6' );
$res = task_par_key('nom-040506!complexe' );
$tid[9] = $res->{'TASK_ID'};
isnt ( $tid[8], $tid[9], '__post_run_serial cas 6' );

# /117,121
# Septieme cas: on attend un heureux evennement quand la fin du dernier fils survient.
# On ne *doit* pas terminer.
new_waiting($tid[6],'heureux');
$res = run_par_task($tid[9]);
del_run($res->{'RUN_ID'});
set_task_state($tid[9],'done');
send_event('EOT-nom-040506!complexe');
$res = [ run_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 7');
is ( $res->[0]->{'RUN_EVENT'}, 'EOT-nom-040506!complexe', '__post_run_serial cas 7' );
del_run($res->[0]->{'RUN_ID'});
$task = task_par_id($tid[6]);
&$post_run_serial($task, $infos, $props, 'EOT-nom-040506!complexe', '__post_run_serial cas 7' );
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 7' );
is ( $res->[0]->{'WAITING_EVENT'}, 'heureux', '__post_run_serial cas 7' );
$res = [ run_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 7' );

# /122,127
# Huitieme cas: on recoit l'heureux evennement, et donc on se termine.
send_event('heureux');
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 8' );
$res = [ run_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 8' );
is ( $res->[0]->{'RUN_EVENT'}, 'heureux', '__post_run_serial cas 8' );
del_run($res->[0]->{'RUN_ID'});
$task = task_par_id($tid[6]);
&$post_run_serial($task,$infos,$props,'heureux');
$res = [ waiting_par_task($tid[6]) ];
is ( $#$res, -1, '__post_run_serial cas 8' );
$res = [ run_par_task($tid[6]) ];
is ( $#$res, 0, '__post_run_serial cas 8' );
is ( $res->[0]->{'RUN_EVENT'}, 'end', '__post_run_serial cas 8' );

# __post_run: Rien de transcendant, il ne traite qu'init (pour passer la tache en 'running')
# et end (pour passer la tache en 'done').
my $post_run = \&FDN::Sql::Workflow::Run::__post_run;

# /128,129
$res = [ run_par_task($tid[6]) ];
del_run ( $res->[0]->{'RUN_ID'} );
$task = task_par_id($tid[6]);
&$post_run($task,$infos,$props,'end');
$task = task_par_id($tid[6]);
is ( $task->{'TASK_STATE'}, 'done', '__post_run' );
$res = event_par_event('EOT-serie/Encore');
is ( ref($res), 'HASH', '__post_run' );

# __post_run_parallel: le cas est nettement plus simple (peut-etre parce que je ne
# m'en sers pas en vrai). Sur l'evennement 'init' tous les fils sont lances. Sur un
# evennement quelconque, il ne se passe rien de particulier. Sur une fin de tache,
# on fait __try_end_task.

# /130,131
$props = { serial => 'un' };
$infos = get_task_infos('para');
$tid[10] = launch_task ( 'para', $props );
ok ( $tid[10], '__post_run_parallel prepare' );
$task = task_par_id($tid[10]);
is ( $task->{'TASK_KEY'}, 'para/un', '__post_run_prallel prepare' );

# /132,142
# Premier cas:
# Verifier l'action de 'init': passer la tache a running, lancer les deux fils,
# se mettre en attente de la fin des deux fils.
$task = task_par_id($tid[10]);
$infos = get_task_infos('para');
$props = get_task_properties($tid[10]);
$res = [ run_par_task($tid[10]) ];
del_run($res->[0]->{'RUN_ID'});
&$post_run($task,$infos,$props,'init');
$task = task_par_id($tid[10]);
is ( $task->{'TASK_STATE'}, 'running', '__post_run_parallel cas 1' );
is ( $task->{'TASK_NBCHILD'}, 2, '__post_run_parallel cas 1' );
$res = task_par_key('serie/un');
is ( $res->{'TASK_STATE'}, 'init', '__post_run_parallel cas 1' );
$tid[11] = $res->{'TASK_ID'};
$res = task_par_key('un-0122!complexe');
is ( $res->{'TASK_STATE'}, 'init', '__post_run_parellel cas 1' );
$tid[12] = $res->{'TASK_ID'};
$res = [ run_par_task($tid[11]) ];
is ( $#$res, 0, '__post_run_parallel cas 1' );
is ( $res->[0]->{'RUN_EVENT'}, 'init', '__post_run_parallel cas 1' );
del_run($res->[0]->{'RUN_ID'});
$res = [ run_par_task($tid[12]) ];
is ( $#$res, 0, '__post_run_parallel cas 1' );
is ( $res->[0]->{'RUN_EVENT'}, 'init', '__post_run_parallel cas 1' );
del_run($res->[0]->{'RUN_ID'});
$res = [ waiting_par_task($tid[10]) ];
is ( $#$res, 1, '__post_run_parallel cas 1' );
for my $w ( @$res ) {
	if ( $w->{'WAITING_EVENT'} eq 'EOT-un-0122!complexe' ) {
		ok ( 1, '__post_run_parallel cas 1' );
	} elsif ( $w->{'WAITING_EVENT'} eq 'EOT-serie/un' ) {
		ok ( 1, '__post_run_parallel cas 1' );
	} else {
		is ( $w->{'WAITING_EVENT'}, 'EOT-un-0122!complexe', '__post_run_parallel cas 1' );
	}
}

# __post_run_atomic: pendant qu'on y est, on va faire les tests de __post_run_atomic
# sur le premier fils.

# /143,145
# Premier cas: on envoit 'init' au fils 'complexe' ($tid[12]), il doit passer
# 'running' et creer un RUN sur 'end'.
$task = task_par_id($tid[12]);
$infos = get_task_infos('essai');
$props = get_task_properties($tid[12]);
&$post_run($task,$infos,$props,'init');
$task = task_par_id($tid[12]);
is ( $task->{'TASK_STATE'}, 'running', '__post_run_atomic cas 1' );
$res = [ run_par_task($tid[12]) ];
is ( $#$res, 0, '__post_run_atomic cas 1' );
is ( $res->[0]->{'RUN_EVENT'}, 'end', '__post_run_atomic cas 1' );
del_run($res->[0]->{'RUN_ID'});

# /146,151
# Deuxieme cas: on envoit 'end' au fils 'complexe' ($tid[12]), il doit passer
# 'done', le pere ($tid[10]) doit avoir un RUN sur 'EOT-un-0122!complexe'. Il ne doit
# plus attendre qu'un seul fils ('EOT-serie/un').
$task = task_par_id($tid[12]);
$infos = get_task_infos('essai');
$props = get_task_properties($tid[12]);
&$post_run($task,$infos,$props,'end');
$task = task_par_id($tid[12]);
is ( $task->{'TASK_STATE'}, 'done', '__post_run_atomic cas 2' );
$res = event_par_event('EOT-un-0122!complexe');
is ( ref($res), 'HASH', '__post_run_atomic cas 2' );
$res = [ waiting_par_task($tid[10]) ];
is ( $#$res, 0, '__post_run_atomic cas 2' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-serie/un', '__post_run_atomic cas 2' );
$res = [ run_par_task($tid[10]) ];
is ( $#$res, 0, '__post_run_atomic cas 2' );
is ( $res->[0]->{'RUN_EVENT'}, 'EOT-un-0122!complexe', '__post_run_atomic cas 2' );
del_run($res->[0]->{'RUN_ID'});

# On revient a __post_run_parallel.

# /152,155
# Deuxieme cas: on envoit 'EOT-un-0122!complexe' au pere ($tid[10]). Il doit rester
# 'runnin'. Il doit rester en attente de son deuxieme fils.
$task = task_par_id($tid[10]);
$infos = get_task_infos('para');
$props = get_task_properties($tid[10]);
&$post_run($task,$infos,$props,'EOT-un-0122!complexe');
$task = task_par_id($tid[10]);
is ( $task->{'TASK_STATE'}, 'running', '__post_run_parallel cas 2' );
$res = [ waiting_par_task($tid[10]) ];
is ( $#$res, 0, '__post_run_parallel cas 2' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-serie/un', '__post_run_parallel cas 2' );
$res = [ run_par_task($tid[10]) ];
is ( $#$res, -1, '__post_run_parallel cas 2' );

# /156,159
# Troisieme cas: on envoit un evennement sans rapport, il ne doit rien se passer.
&$post_run($task,$infos,$props,'sans-rapport');
$task = task_par_id($tid[10]);
is ( $task->{'TASK_STATE'}, 'running', '__post_run_parallel cas 3' );
$res = [ waiting_par_task($tid[10]) ];
is ( $#$res, 0, '__post_run_parallel cas 3' );
is ( $res->[0]->{'WAITING_EVENT'}, 'EOT-serie/un', '__post_run_parallel cas 3' );
$res = [ run_par_task($tid[10]) ];
is ( $#$res, -1, '__post_run_parallel cas 3' );

# /160,164
# Quatrieme cas: on termine le fils 'serie' ($tid[11]), on envoit 'EOT-serie/un'
# au pere ($tid[10]). Il doit recuperer un RUN sur 'EOT-serie/un'.
set_task_state($tid[11],'running');
$task = task_par_id($tid[11]);
$infos = get_task_infos('serie');
$props = get_task_properties($tid[11]);
&$post_run($task,$infos,$props,'end');
$task = task_par_id($tid[11]);
is ( $task->{'TASK_STATE'}, 'done', '__post_run_parallel cas 4' );
$res = event_par_event('EOT-serie/un');
is ( ref($res), 'HASH', '__post_run_parallel cas 4' );
$res = [ waiting_par_task($tid[10]) ];
is ( $#$res, -1, '__post_run_parallel cas 4' );
$res = [ run_par_task($tid[10]) ];
is ( $#$res, 0, '__post_run_parallel cas 4' );
is ( $res->[0]->{'RUN_EVENT'}, 'EOT-serie/un', '__post_run_parallel cas 4' );
del_run($res->[0]->{'RUN_ID'});

# /165,167
# Cinquieme cas: on envoit 'EOT-serie/un' au pere, il doit recuperer un RUN sur 'end'.
$task = task_par_id($tid[10]);
$infos = get_task_infos('para');
$props = get_task_properties($tid[10]);
&$post_run($task,$infos,$props,'EOT-serie/un');
$res = [ waiting_par_task($tid[10]) ];
is ( $#$res, -1, '__post_run_parallel cas 5' );
$res = [ run_par_task($tid[10]) ];
is ( $#$res, 0, '__post_run_parallel cas 5' );
is ( $res->[0]->{'RUN_EVENT'}, 'end', '__post_run_parallel cas 5' );
del_run($res->[0]->{'RUN_ID'});

# /168,170
# Sixieme cas: on envoit 'end' au pere, il doit passer a 'done'
$task = task_par_id($tid[10]);
$infos = get_task_infos('para');
$props = get_task_properties($tid[10]);
&$post_run($task,$infos,$props,'end');
$task = task_par_id($tid[10]);
is ( $task->{'TASK_STATE'}, 'done', '__post_run_parallel cas 5' );
$res = [ waiting_par_task($tid[10]) ];
is ( $#$res, -1, '__post_run_parallel cas 5' );
$res = [ run_par_task($tid[10]) ];
is ( $#$res, -1, '__post_run_parallel cas 5' );
# Before 28 + (/170) -> 198


####
#### Top level functions -> 70
####

# /1,40
# purge_task: facile, il ne doit rien rester!
for my $tid ( @tid ) {
	# 5 fois seulement: les taches filles sont enlevees par les taches meres.
	$task = task_par_id($tid);
	next if ! defined $task;
	purge_task($tid);
	$res = task_par_id($tid);
	is ( $res, undef, 'purge_task' );
	$res = task_par_key($task->{'TASK_KEY'});
	is ( $res, undef, 'purge_task' );
	$res = get_task_properties($tid);
	is_deeply ( $res, {}, 'purge_task' );
	$res = waiting_par_task($tid);
	is ( $res, undef, 'purge_task' );
	$res = run_par_task($tid);
	is ( $res, undef, 'purge_task' );
	$res = event_par_event('EOT-'.$task->{'TASK_KEY'});
	is ( $res, undef, 'purge_task' );
	$res = taskself_par_parent($tid);
	is ( $res, undef, 'purge_task' );
	$res = taskself_par_fils($tid);
	is ( $res, undef, 'purge_task' );
}

# run_task: nettement plus casse pieds!
# On se prepare 4 taches, qu'on va faire tourner. Une de chaque type.

my $tid1 = launch_task('essai',{
	nom => "Benjamin",
	telephone => 42,
	type => 'essai'
	});
# Va faire deux RUNs: 'init' et 'end'.

my $tid2 = launch_task('complexe',{
	nom => "Benjamin",
	telephone => "01000000",
	type => 'complexe'
	});
# Va faire deux RUNs: 'init' et 'end'.

my $tid3 = launch_task('serie',{
	nom => "Essai suivant",
	telephone => "012345678",
	type => 'serie'
	});
# Va faire 4 runs: 'init', 2 EOT, 'end'.
# Plus 2 fils de 2 runs
# -> 8 runs

my $tid4 = launch_task('para', {
	pouet => "rien",
	serial => 42,
	type => 'para'
	});
# Va faire 4 runs: 'init', 2 EOT, plus  'end'.
# plus 1 fils de 2 runs
# plus 1 fils de 8 runs (type 'serie')
# -> 14 runs

# 41,66
my $run;
lock_run();
$run = run_par_state('free');
while ( $run ) {
	info "Je traite le run ".$run->{'RUN_ID'}." il porte sur la tache".
		$run->{'TASK_ID'}."\n";
	set_run_state($run->{'RUN_ID'},'busy');
	unlock_run();
	$res = run_task($run);
	isnt ( $res, undef, 'run_task' );
	lock_run();
	$run = run_par_state('free');
}
info "Rien de plus a faire tourner";

# /67,70
$task = task_par_id($tid1);
is ( $task->{'TASK_STATE'}, 'done', 'run_task' );
$task = task_par_id($tid2);
is ( $task->{'TASK_STATE'}, 'done', 'run_task' );
$task = task_par_id($tid3);
is ( $task->{'TASK_STATE'}, 'done', 'run_task' );
$task = task_par_id($tid4);
is ( $task->{'TASK_STATE'}, 'done', 'run_task' );

purge_task($tid1);
purge_task($tid2);
purge_task($tid3);
purge_task($tid4);
# Before 198 + (/70) -> 268

####
#### Event handling -> 208
####

$OUTPUT_AUTOFLUSH=1;
@tid = ();

# Quelques routines utilitaires:
sub add_to_test($) {
	my $builder = Test::More->builder;
	my $n = $builder->current_test;
	$builder->current_test($n+$_[0]);
}
sub use_db($) {
	$dbh = do_connect($_[0]);
	workflow_select_db($_[0]);
}

sub joueur($$$) {
	my ( $wait, $skip, $cas ) = @_;
	if ( ! fork() ) {
		# Voici le joueur
		use_db('test');
		add_to_test($skip);
		sleep(1+$wait);
		$res = wait_event($tid,"ev-cas$cas");
		ok ( defined $res, "wait_event cas $cas" );
		print "# Joueur a demande waiting sur ev-cas$cas\n";
		do_disconnect('test');
		exit(0);
	}
}

sub analyste($$$) {
	my ( $wait, $skip, $cas ) = @_;
	if ( ! fork() ) {
		# Voici l'analyste
		use_db('test');
		add_to_test(4);
		sleep(1+$wait);
		$res = send_event("ev-cas$cas");
		is ( $res, TRUE, "send_event cas $cas" );
		print "# Analyste a produit evennement ev-cas$cas\n";
		do_disconnect('test');
		exit(0);
	}
}

sub is_waiting($) {
	my ( $cas ) = @_;
	$res = waiting_par_event("ev-cas$cas");
	is ( ref($res), 'HASH', "event cas $cas");
	is ( $res->{'TASK_ID'}, $tid, "event cas $cas");
	is ( $res->{'WAITING_EVENT'}, "ev-cas$cas", "event cas $cas");
	$res = run_par_task($tid);
	is ( $res, undef, "event cas $cas");
	$res = event_par_event("ev-cas$cas");
	is ( $res, undef, "event cas $cas");
}

sub is_to_run($) {
	my ( $cas ) = @_;
	$res = waiting_par_event("ev-cas$cas");
	is ( $res, undef, "event cas $cas" );
	$res = run_par_task($tid);
	is ( ref($res), 'HASH', "event cas $cas" );
	is ( $res->{'TASK_ID'}, $tid, "event cas $cas" );
	is ( $res->{'RUN_EVENT'}, "ev-cas$cas", "event cas $cas" );
	$res = event_par_event("ev-cas$cas");
	is ( ref($res), 'HASH', "event cas $cas");
}

sub no_wait_run($) {
	my ( $cas ) = @_;
	$res = waiting_par_event("ev-cas$cas");
	is ( $res, undef, "event cas $cas");
	$res = run_par_task($tid);
	is ( $res, undef, "event cas $cas");
	$res = event_par_event("ev-cas$cas");
	is ( ref($res), 'HASH', "event cas $cas" );
}

sub virgin_task($) {
	my ( $cas ) = @_;
	$tid = launch_task('essai', { nom => "cas$cas", telephone=> '010000' } );
	$res = run_par_task($tid);
	is ( $res->{'RUN_EVENT'}, 'init', 'supression d init');
	del_run($res->{'RUN_ID'});
}

# 3 process sont utilises:
# - maitre (moi)
# - joueur (fils 1)
# - analyste (fils 2)
# Le joueur joue une tache qui demande sa mise en attente d'un evennement
# (wait_event).
# L'analyste decouvre dans l'environnement exterieur l'evennement tant attendu.
# Le maitre sert a controler le jeu.


# Cas 1:
# Le joueur passe avant l'analyste.
# /1
virgin_task(1);
do_disconnect($dbh);
joueur(0,0,1);
analyste(2,6,1); # Il faut passer le jouer + 5 is_waiting
use_db('test');
# Il se passe une seconde, le jouer et l'analyste dorment.

# /2
print "# On attend le joueur.\n";
add_to_test(1);
sleep(2);
# Le joueur a fait son travail.

# /3,7
is_waiting(1); # 5 tests

# /8
print "# On attend l'analyste.\n";
add_to_test(1);
sleep(2);
# L'analyste a du creer l'evennement, degager l'attente et creer un run.

# /9,13
is_to_run(1); # 5 tests




# Cas 2:
# L'analyste passe avant le joueur.
# /14
virgin_task(2);
do_disconnect($dbh);
analyste(0,0,2);
joueur(2,4,2); # Il faut passer l'analyste + 4 no_wait_run
use_db('test');

# /15
print "# On attend l'analyste\n";
add_to_test(1);
sleep(2);
# L'analuste a du creer l'evennement et c'est tout.

# /16,18
no_wait_run(2); # 3 tests

# /19
print "# On attend le joueur\n";
add_to_test(1);
sleep(2);
# Le joueur a du creer le run.

# /20,24
is_to_run(2); # 5 tests



# Cas 3: ordre indetermine
# /25
virgin_task(3);
do_disconnect($dbh);
analyste(0,0,3);
joueur(0,1,3); # Le joueur passe le test, risque de desordre (pas grave).
use_db('test');

# /26,27
print "# On attend le jouer et l'analyste\n";
add_to_test(2);
sleep(2);

# /28,32
is_to_run(3);



# /33,120
# Cas 4->14 en force brute.
for my $n ( 4..14 ) {
	virgin_task($n);
	do_disconnect($dbh);
	analyste(-1,0,$n);
	joueur(-1,0,$n);
	use_db('test');
	add_to_test(2);
	sleep(1);
	is_to_run($n); # 5 tests
}

# Cas 15->25 force nettement plus brute!
# /121,131
for my $n ( 15..25 ) {
	virgin_task($n);
	$tid[$n]=$tid;
}
do_disconnect($dbh);
# /132,153
for my $n ( 15..25 ) {
	$tid = $tid[$n];
	analyste(-1,0,$n);
	joueur(-1,0,$n);
}
add_to_test(22);
# /154,208
sleep(5);
for my $n ( 15..25 ) {
	$tid = $tid[$n];
	is_to_run($n);
}
# Before 268 + (/208) -> 476

END {
	workflow_close_db();
}

exit(0);

# Ici les fonctions utiles pour les taches elles-meme.

#sub faire_telephone() {
#	return '0140280941';
#}
#
#sub calculer_la_clef($$) {
#	my ( $name,$vars ) = @_;
#	return $vars->{'nom'}.'-'.$vars->{'telephone'}.'!'.$name;
#}
#
#sub faire_un_essai($$$) {
#	my ( $props, $event, $date ) = @_;
#	my $type;
#	if ( defined $props->{'type'} ) {
#		$type = $props->{'type'};
#	} else {
#		$type = 'sub of '.$props->{'ptid'};
#	}
#	my $tid = $props->{'tid'};
#	print "# Je ($tid type $type) recois l'evenement $event en date de $date\n";
#	return '';
#}


